"""a wrapper for memuc.exe as a library to control virual machines"""

import logging
from os.path import join
from typing import Union

from ._constants import WINREG_EN
from .exceptions import PyMemucError


class PyMemuc:
    """A class to interact with the memuc.exe command line tool to control virtual machines.

    :param memuc_path: Path to memuc.exe. Set to None for autodetect.
    :type memuc_path: str, optional
    :param debug: Enable debug mode, defaults to False
    :type debug: bool, optional
    """

    # pylint: disable=import-outside-toplevel

    from ._command import (
        change_gps_vm,
        connect_internet_vm,
        create_app_shortcut_vm,
        disconnect_internet_vm,
        execute_command_vm,
        get_app_info_list_vm,
        get_public_ip_vm,
        input_text_vm,
        install_apk_vm,
        rotate_window_vm,
        send_adb_command_vm,
        set_accelerometer_vm,
        sort_out_all_vm,
        start_app_vm,
        stop_app_vm,
        trigger_keystroke_vm,
        trigger_shake_vm,
        uninstall_apk_vm,
        zoom_in_vm,
        zoom_out_vm,
    )
    from ._control import reboot_vm, start_vm, stop_all_vm, stop_vm
    from ._manage import (
        clone_vm,
        create_vm,
        delete_vm,
        export_vm,
        get_configuration_vm,
        import_vm,
        list_vm_info,
        randomize_vm,
        rename_vm,
        set_configuration_vm,
        vm_is_running,
    )
    from ._memuc import _get_memu_top_level, check_task_status, memuc_run

    def __init__(self, memuc_path: Union[str, None] = None, debug=False) -> None:
        """initialize the class, automatically finding memuc.exe if windows registry is supported,
        otherwise a path must be specified"""
        self.debug = debug
        self.logger = self._configure_logger()
        self.logger.debug("PyMemuc: Debug mode enabled")
        if WINREG_EN:
            self.memuc_path: str = join(self._get_memu_top_level(), "memuc.exe")
        elif memuc_path is not None:
            self.memuc_path: str = memuc_path
        else:
            raise PyMemucError(
                "Windows Registry is not supported on this platform,"
                + " you must specify the path to memuc.exe manually"
            )

    def _configure_logger(self):
        """Configure the logger for the class"""
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG if self.debug else logging.INFO)

        # Create a handler for console output
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        console_handler.setFormatter(formatter)

        # Add the handler to the logger
        logger.addHandler(console_handler)

        return logger



"""This module contains functions for directly interacting with memuc.exe."""
from contextlib import suppress
from os.path import join, normpath
from subprocess import PIPE, STDOUT, CalledProcessError, Popen, TimeoutExpired
from typing import TYPE_CHECKING, Tuple

from ._constants import WIN32, WINREG_EN
from .exceptions import PyMemucError, PyMemucException, PyMemucTimeoutExpired

if WINREG_EN:
    # pylint: disable=import-error
    from winreg import HKEY_LOCAL_MACHINE, ConnectRegistry, OpenKey, QueryValueEx

if WIN32:
    from subprocess import STARTF_USESHOWWINDOW, STARTUPINFO, SW_HIDE

    ST_INFO = STARTUPINFO()
    ST_INFO.dwFlags = STARTF_USESHOWWINDOW
    ST_INFO.wShowWindow = SW_HIDE
else:
    ST_INFO = None

if TYPE_CHECKING:
    from pymemuc import PyMemuc


@staticmethod
def _get_memu_top_level() -> str:
    """locate the path of the memu directory using windows registry keys

    :return: the path of the memu directory
    :rtype: str
    :raises PyMemucError: an error if memu is not installed
    """
    if not WINREG_EN:
        raise PyMemucError("Windows Registry is not supported on this platform")
    for key in [  # keys to search for memu
        r"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\MEmu",
        r"SOFTWARE\WOW6432Node\Microsoft\Windows\CurrentVersion\Uninstall\MEmu",
    ]:
        with suppress(FileNotFoundError):
            akey = OpenKey(ConnectRegistry(None, HKEY_LOCAL_MACHINE), key)
            return str(join(normpath(QueryValueEx(akey, "InstallLocation")[0]), "Memu"))
    raise PyMemucError("MEmuc not found, is it installed?")


def memuc_run(
    self: "PyMemuc", args: list[str], non_blocking=False, timeout=None
) -> Tuple[int, str]:
    """run a command with memuc.exe.
    Memuc can support non-blocking commands, returning a task id.
    A timeout can be specified if memuc is expected to hang,
    but this will not work with non-blocking commands.

    :param args: a list of arguments to pass to memuc.exe
    :type args: list[str]
    :param non_blocking: whether to run the command in the background. Defaults to False.
    :type non_blocking: bool, optional
    :param timeout: the timeout in seconds. Defaults to None for no timeout.
    :type timeout: int, optional
    :return: the return code and the output of the command
    :rtype: tuple[int, str]
    :raises PyMemucError: an error if the command failed
    :raises PyMemucTimeoutExpired: an error if the command timed out
    """
    # sourcery skip: extract-method
    args.insert(0, self.memuc_path)
    if timeout is not None and non_blocking:
        raise PyMemucException("Cannot use timeout and non_blocking at the same time")
    if non_blocking:
        args += "-t"
    self.logger.debug("pymemuc._memuc.memuc_run:")
    self.logger.debug(f"\tCommand: \"{' '.join(args)}\"")
    try:
        with Popen(
            args,
            shell=False,
            stdout=PIPE,
            stderr=STDOUT,
            close_fds=True,
            universal_newlines=True,
        ) as process:
            try:
                result, _ = process.communicate(timeout=timeout)
            except TimeoutExpired as err:
                process.kill()
                result, _ = process.communicate()
                raise PyMemucTimeoutExpired(err) from err
            if lines := result.splitlines():
                self.logger.debug(f"\tOutput: {lines.pop(0)}")
                for line in lines:
                    self.logger.debug(f"\t\t{line}")
            return (0, result)
    except CalledProcessError as err:
        raise PyMemucError(err) from err


# TODO: add output parsing
def check_task_status(self: "PyMemuc", task_id: str) -> Tuple[int, str]:
    """Check the status of a task

    :param task_id: Asynchronous task ID
    :type task_id: str
    :return: the return code and the output of the command.
    :rtype: tuple[int, str]
    """
    return self.memuc_run(["taskstatus", task_id])
